//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Ds9.cc:	--- implementation of Ds9 class
//
//	CHANGE HISTORY:
//		UL, 1999-05-06: 1.0.0: - first version; uses Ds9-1.8-beta
//		UL, 1999-05-19: 1.1.0: - updated for Ds9-1.8.2-beta
//							   - set coordinate system for region markers
//								 to fk5 before reading region data
//		UL, 1999-08-17: 1.2.0: - bug fix in Ds9::regionSelection()
//		UL, 1999-09-20: 1.2.1: - do no longer check return of of 'system'
//								 used to bring up Ds9 (causes problem at
//								 Leceister)
//							   - inquire regions from Ds9 in SAOtng format
//								 for ingesiton by
//								 ExprTranslator::toSelectExpression in
//							     selectlib package
//		UL, 1999-10-27: 1.2.2: - brought in line with ds9-1.8.6.9
//								 (uses no longer '-' in front of options
//								 of XPA access points)
//		UL, 1999-10-28: 1.2.3: - in Ds9::receive(): moved first
//								 switchSignalHanlders() statement to the
//								 beginning of the function; otherwise
//								 xmgr interferes with the xmmselect-ds9
//								 pipe
//		UL, 1999-12-17: 1.3.0: - brought in line with selectlib-4.23
//		UL, 2000-01-05: 1.4.0: - only change settings (new colormap/log scale)
//								 when bringing up new ds9
//								 (fixes SPR 1034 : should not reset ds9
//								 settings)
//		UL, 2000-01-19: 1.5.0: - implemented Ds9::displayRegion()
//		UL, 2000-02-02: 1.6.0: - write "linear" keyword to Ds9 region file only
//								 if marker point coordinates are in WCS
//		UL, 2000-03-14: 1.7.0: - no longer do a "regions deleteall" before
//								 displaying a new one
//		UL, 2000-07-21: 1.8.0: - adapted Ds9::toRegion()/<<operator to
//								 ds9 region file format 3.0
//		UL, 2000-12-22: 1.8.1: - trivial change in Ds9::toRegion() to make
//								 optimizer happy
//		UL, 2001-08-17: 1.8.2: - fixed problem in writing of Ds9 region
//								 file [used to use operator<<(ostream &,
//								 const Point<T> &) from selectlib but
//								 its output format has changed]
//		UL, 2002-01-28: 1.8.3: - toRegion(): added support for elliptical
//								 annuli
//		UL, 2002-01-30: 1.8.4: - toRegion(): adapted to account for changed
//								 annulus region format in ds9 2.1b4
//		UL, 2002-02-19: 1.9.0: - adjusted command for colormap selection
//								 to 2.1b5 (cmap heat)
//		UL, 2002-07-11: 2.0.0: - Ds9::toRegion: always return region with
//								 two components (included + globally exluded)
//		UL, 2002-09-03: 2.1.0: - new stopViewer argument to constructor
//		UL, 2002-09-18: 2.2.0: - implementation of various new functionalities
//		UL, 2002-11-11: 2.3.0: - added Ds9::deleteRegions()
//		UL, 2002-11-13: 2.4.0: - correctly extract global marker color from
//								 global attributes in Ds9::toRegion()
//							   - correctly treat include status of components
//								 operator<<(ostream, Region)
//		UL, 2002-11-26: 2.5.0: - modifications in Ds9::toRegion() to support
//								 recognition of ds9 background regions
//		UL, 2002-12-18: 2.6.0: - proper recognition of elliptical annuli in
//								 ds9>=2.3b1
//		UL, 2003-01-08: 2.6.1: - fixed bug in Ds9::toRegion() [multiple
//								 consecutive ellipses could trigger crash]
//		UL, 2003-02-21: 2.6.2: - fixed bug in Ds9::toRegion() which caused
//								 region text attributes not to be recognized
//		UL, 2003-05-23: 2.7.0: - make use of ShapeBasics::toColor/toString()
//								 instead of hardcoded color strings
//              AI, 2006-03-01: 2.7.1: - switchSignalHandlers no longer needed, removed
//
//	AUTHOR:
//		U. Lammers (UL), Thu May  6 11:20:00 MET DST 1999
//
#include <fstream>
#include <iostream>
//#include <streambuf>	//	ios_base::sync_with_stdio
#include <cstdlib>
#include <cctype>
#include <cstdio>	//	tmpnam()
#include <csignal>
#include <unistd.h>
#include "Ds9.h"
#include "SasImageDisplayer.h"
#include "Region.h"
#include "ASCregion.h"
#include "errstr.h"
#include "stringutil.h"
#include "common.h"

using namespace std;

const std::string Ds9::Xpaset("xpaset");
const std::string Ds9::Xpaget("xpaget");
const std::string Ds9::TITENV("LOGNAME");




//	construct Ds9 wrapper object
Ds9::Ds9(bool stopViewer) : ImageDisplayer(), _DS9("ds9")
{
	setName();

	ios::sync_with_stdio();	//	we use popen/pclose
	if (isThere()) {
		_title = "Ds9";
		char *p = std::getenv(TITENV.c_str());
		if (p)
			_title += "-" + std::string(p);

		if (!isRunning()) {
			if (stopViewer)
				return;
			//	is's not running - try to bring it up
			execBackground(_DS9 + " -title " + _title);
			//	wait till it comes up and is receptive
			unsigned nChecks = 0;
			while (!isRunning() && nChecks++ <= 20)
				sleep(1);
			if (nChecks <= 20) {
				//	ok - up and receptive; set nice colormap, zoom in a bit
				//	and switch to scale log
				send("orient none");
				send("zoom to 2");
				send("scale log");
				send("cmap heat");

				return;
			}
		} else {
			//	is already running
			if (stopViewer)
				down();
			return;
		}
	}

	errstr<<"Could not launch Ds9. It is either not "
			"installed or inaccessible via the current PATH."
		  <<Sas::error("UnableToLaunchProgram");
} // Ds9::Ds9



void Ds9::flipX() const
{
	std::vector<std::string> v(receive("orient"));

	//
	//	depending on what the current orientation is we need to
	//	do send different commands to achieve the desired result:
	//
	//		current orientation:	new orientation:
	//		--------------------    ----------------
	//		none					x
	//		x						none
	//		y						xy
	//		xy						y
	//
	if (v.size() == 1)
		if (v[0] == "none")
			send("orient x");
		else if (v[0] == "x")
			send("orient none");
		else if (v[0] == "y")
			send("orient xy");
		else if (v[0] == "xy")
			send("orient y");
		else
			errstr<<"Current image orientation is `"<<v[0]<<"\' - "
					"not understood"
				  <<Sas::warning("UnknownOrientation");
	else
		errstr<<"Could not get actual orientation of the image - not flipped"
			  <<Sas::warning("UnknownOrientation");
} // Ds9::flipX()


void Ds9::flipY() const
{
	std::vector<std::string> v(receive("orient"));

	//
	//	depending on what the current orientation is we need to
	//	do send different commands to achieve the desired result:
	//
	//		current orientation:	new orientation:
	//		--------------------    ----------------
	//		none					x
	//		x						none
	//		y						xy
	//		xy						y
	//
	if (v.size() == 1)
		if (v[0] == "none")
			send("orient y");
		else if (v[0] == "x")
			send("orient xy");
		else if (v[0] == "y")
			send("orient none");
		else if (v[0] == "xy")
			send("orient x");
		else
			errstr<<"Current image orientation is `"<<v[0]<<"\' - "
					"not understood"
				  <<Sas::warning("UnknownOrientation");
	else
		errstr<<"Could not get actual orientation of the image - not flipped"
			  <<Sas::warning("UnknownOrientation");
} // Ds9::flipY()


void Ds9::zoomIn(ZoomFactor) const
{
	errstr<<"Zooming in not yet implemented - nothing done"
		  <<Sas::warning("NotYetImplemented");
} // Ds9::zoomIn()


void Ds9::zoomOut(ZoomFactor) const
{
	errstr<<"Zooming out not yet implemented - nothing done"
		  <<Sas::warning("NotYetImplemented");
} // Ds9::zoomOut()


void Ds9::displayImage(const std::string &dataSetName) const
{
	send("file " + dataSetName);
} // Ds9::displayImage()


void Ds9::displayRegion(const std::string &regionTable,
						CoordType coordType) const
{
	InASCregionTable inRegionTable(regionTable);
	GeoRegion region;

	inRegionTable>>region;	//	read region from table
	ImageDisplayer::Region r(region, coordType);

	displayRegion(r);
} // Ds9::displayRegion()


void Ds9::displayRegion(const Region &region) const
{
	char *regionFile = std::tmpnam(0);
	std::ofstream of(regionFile);
	if (!of)
		errstr<<"Cannot open region file `"<<regionFile<<"' for write"
			  <<Sas::error("UnableToCreateFile");

	of<<ImageDisplayer::Region(region)<<endl;
	of.close();

//	send("regions deleteall");
//	send("zoom to 1");
	send("regions file " + (std::string)regionFile);

	unlink(regionFile);
} // Ds9::displayRegion()


std::vector<std::string> Ds9::regionSelectionAsString(CoordType coordType)
const
{
	string coordTypeStr = "-system image";
	if (coordType == LINEAR)
		coordTypeStr = "-system physical";
	else if (coordType == FK5)
		coordTypeStr = "-sky fk5 -skyformat degrees";

	//	get raw image coordinates back in ds9 format
	return receive("regions " + coordTypeStr + " -format ds9");
} // Ds9::regionSelection()


::GeoRegion Ds9::regionSelection(ImageDisplayer::CoordType coordType,
								 const std::string &ignoreMarkerColor) const
{
	return toRegion(regionSelectionAsString(coordType), ignoreMarkerColor);
} // Ds9::regionSelection()


void Ds9::deleteRegions() const
{
	send("regions deleteall");
} // Ds9::deleteRegions()


void Ds9::down() const
{
	send("exit");
} // Ds9::down()


Ds9::~Ds9()
{
	if (isRunning())
		down();
} // Ds9::~Ds9()


void Ds9::send(const std::string &command) const
{
	std::string commandLine = Xpaset + " " + _title;
	FILE *xpasetPipe = popen(commandLine.c_str(), "w");
	if (!xpasetPipe ||
		std::fprintf(xpasetPipe, "%s\n", command.c_str()) <= 0)
		errstr<<"Could not open or write on pipe to `"<<Xpaset
			  <<"' - please note that `"<<Xpaset<<"' must be accessible via"
				" the PATH. Could this be the problem?"
			  <<Sas::warning("UnableToSendCommand");

	//	see above why whis is necessay
	pclose(xpasetPipe);
} // Ds9::send()


std::vector<std::string> Ds9::receive(const std::string &what) const
{
	std::vector<std::string> response;
	std::string commandLine = Xpaget + " " + _title + " " + what + " 2>&1";
	FILE *xpagetPipe = popen(commandLine.c_str(), "r");
	if (xpagetPipe) {
		char s[2048];
		while (std::fgets(s, sizeof(s)-1, xpagetPipe)) {
			if (std::strlen(s) && s[std::strlen(s)-1] == '\n')
				s[std::strlen(s)-1] = '\0';
			if (s[0] != '#')	//	comments
				response.push_back(s);
		}
	} else
		errstr<<"Could not open or write on pipe to `"<<Xpaget
			  <<"' - please note that `"<<Xpaget<<"' must be accessible via"
				" the PATH. Could this be the problem?"
			  <<Sas::warning("UnableToReceive");

	//	see above why whis is necessay
	pclose(xpagetPipe);

	return response;
} // Ds9::receive()


bool Ds9::isRunning() const
{
	std::vector<std::string> v(receive(""));

	if (v.size() == 0)
		return true;

	v = breakup(v[0], " ");

	return v[0] != "XPA$ERROR";
} // Ds9::isRunning()


bool Ds9::isThere() const
{
	return true;
} // Ds9::isThere()


void Ds9::execBackground(const std::string &command) const
{
	std::string comm = command + " >/dev/null 2>&1 &";
	std::system(comm.c_str());
} // Ds9::execBackground()


void Ds9::setName() {
	char *p = getenv(SAS_IMAGE_VIEWER_ENV_NAME);
	if (p)
		_DS9 = p;
} // Ds9::setName()


::GeoRegion Ds9::toRegion(const std::vector<std::string> &regionSpec,
						  const std::string &ignoreMarkerColor)
{
	bool ignoreMarkers = !ignoreMarkerColor.empty();
	ShapeBasics::ShapeColor ignCol = ShapeBasics::BLACK;
	if (ignoreMarkers)
		ignCol = ShapeBasics::toColor(ignoreMarkerColor);

	std::string knownRegions = "point circle annulus ellipse polygon box";
	ShapeBasics::ShapeColor stdColor = ShapeBasics::WHITE;

	::GeoRegion		region;
	region.push_back(Component());	//	included component
	region.push_back(Component());	//	excluded component
	region[1].exclude();
	
	for (unsigned n=0; n<regionSpec.size(); ++n) {
		std::string s(trim(regionSpec[n], true));	//	make a copy
		if (s.empty() || s[0] == '#')
			continue;	//	skip empty + comment lines (starting with '#')

		//	chop off everything after '#'
		string::size_type pos;
		string comment;
		if ((pos = s.find("#")) != std::string::npos) {
			comment=trim(s.substr(pos+1), true);
			s=trim(s.substr(0, pos), true);
		}

		if (s.length()<=5 || lowercase(s.substr(0, 6)) == "global") {
			if ((pos = s.find("color=")) != string::npos) {
				s = trim(s.substr(pos+6));
				vector<string> v(breakup(s, " "));
				if (v.size())
					stdColor = ShapeBasics::toColor(v[0]);
			}
			continue;	//	not interested in other global attributes
		}

		bool linear = false;
		if ((pos = s.find(";")) != std::string::npos) {
			linear = lowercase(trim(s.substr(0, pos))) == "linear";
			s = trim(s.substr(pos+1), true);
		}

		//
		//	now s should contain just the marker
		//	now break on "(), &" to get region arguments
		//	'&' is used to form intersections, e.g.
		//	ellipse(9,11,3,14,0) & !ellipse(9,11,3.3,49.33,0)
		//
		std::vector<std::string> fields(breakup(s, "(, )&"));
		if (fields.size() < 3)
			continue;

		#define THROWINVALID(minsize)	\
			if (fields.size() < minsize) { throw GeoRegion::InvalidRegion(); }

		std::string regionType = lowercase(fields[0]);
		bool included = regionType[0] != '-';

		if (!included)
			regionType.erase(0, 1);

		if (knownRegions.find(regionType) == std::string::npos)
			continue;

		Component *comp = &region[included ? 0 : 1];
		FShape *shape;
		if (regionType != "polygon") 
		  {
		    FPoint mid(toNumber<float>(fields[1]),
			       toNumber<float>(fields[2]));
		    if (regionType == "point")
		      shape = new FShape(mid);
		    else if (regionType == "circle") 
		      {
			THROWINVALID(4);
			shape = new FCircle(mid, toNumber<float>(fields[3]));
			} 
		    else if (regionType == "annulus") 
		      {
			THROWINVALID(5);
			unsigned rInnerIndx = 3,
			  rOuterIndx = 4;
			if (toNumber<float>(fields[3]) == 0. && fields.size() > 5)
			  //	this is a multi-annulus region with the a 0-radius
			  //	of innermost ring - take second circle as
			  //	defining inner annulus radius
			  ++rInnerIndx, ++rOuterIndx;
			
			shape = new FAnnulus(mid,
					     toNumber<float>(fields[rInnerIndx]),
					     toNumber<float>(fields[rOuterIndx]));
		      } 
		    else if (regionType == "ellipse") 
		      {
			//	check if this is an elliptical annulus; in ds9 this
			//	is defined in the weird way
			//		innerEllipse
			//		outerEllipse & !innerEllipe
			if (comp->size() >= 1 &&
			    comp->back()->type() == ShapeBasics::ELLIPSE) 
			  {
			    //	last registered shape in component was ellipse
			    FShape *e = comp->back();
			    if (fields.size() == 12 &&
				lowercase(fields[6]) == "!ellipse") 
			      {
				//	check if negated ellipse is the same as the
				//	last registered one
				float x = toNumber<float>(fields[7]),
				  y = toNumber<float>(fields[8]),
				  ri = toNumber<float>(fields[9]),
				  ro = toNumber<float>(fields[10]),
				  alpha = toNumber<float>(fields[11]);
				if (x == e->refPoint().x() && y == e->refPoint().y() &&
				    ri == e->radii()[0] && ro == e->radii()[1] &&
				    alpha == e->rotAngles()[0]) 
				  {
				    //	yes, this is an elliptical annulus; remove
				    //	last registered ellipse
				    delete e;
				    shape = new FElliptAnnulus(mid, ri, ro,
							       toNumber<float>(fields[3]),
							       toNumber<float>(fields[4]),
							       toNumber<float>(fields[5]),
							       alpha, alpha);
				    comp->pop_back();						
				    continue;
				  }
			      }
			  }
			
			THROWINVALID(6);
			if (fields.size() == 6)
			  shape = new FEllipse(mid, toNumber<float>(fields[3]),
					       toNumber<float>(fields[4]),
					       toNumber<float>(fields[5]));
			else
			  //	this is an elliptical annulus
			  shape = new FElliptAnnulus(mid,
						     toNumber<float>(fields[3]),
						     toNumber<float>(fields[4]),
						     toNumber<float>(fields[5]),
						     toNumber<float>(fields[6]),
						     toNumber<float>(fields.back()),
						     toNumber<float>(fields.back()));
		      } 
		    else if (regionType == "box") 
		      {
			THROWINVALID(6);
			shape = new FRotBox(mid, toNumber<float>(fields[3]),
					    toNumber<float>(fields[4]),
					    toNumber<float>(fields[5]));
		      }
		  } 		
		else 
		  { // polygon
		    vector<float> x, y;
		    unsigned N = fields.size(), i = 1;
		    while (i+2 <= N && !isalpha(fields[i][0])) {
		      x.push_back(toNumber<float>(fields[i]));
		      y.push_back(toNumber<float>(fields[i+1]));
		      i += 2;
		    }
		    shape = new FPolygon(x, y, x.size());
		  }
		shape->color(stdColor);


		//
		//	process optional arguments in comments
		//		color=
		//		text={} | text="bla"
		//		background
		//
		vector<string> v0(quotedBreakup(comment, " ", "\"{"));
		for (unsigned n=0; n<v0.size(); ++n) 
		  {
		    vector<string> v1(quotedBreakup(v0[n], "=", "\"{"));
		    switch (v1.size()) 
		      {
		      case 1:	//	single attribute
			v1[0] = lowercase(v1[0]);
			if (v1[0] == "background")
			  shape->attribute(ShapeBasics::BACKGROUND);
			else if (v1[0] == "source")
			  shape->attribute(ShapeBasics::SOURCE);
			//	ignore all other attributes (for the moment)
			break;
		      case 2: 
			{
			  //	keyword-value pair
			  string key = lowercase(v1[0]),
			    &val = v1[1];
			  if (key=="color")
			    shape->color(ShapeBasics::toColor(val));
			  else if (key=="text") 
			    {
			      string text(unquote(val));
			      if (!text.empty() && text[0] == '{') 
				{
				  text.erase(0, 1);
				  pos = text.find("}");
				  if (pos != std::string::npos)
				    text.erase(pos, 1);
				}
			      shape->label(text);
			    }
			}
			//	ignore the rest for moment - if we want to process more
			//	options this should happen here
			break;
		      default:
			//	unknown attributes or none
			break;
		      }
		  }
		
		//
		//	check if we want to ignore this component
		//
		if (ignoreMarkers && shape->color() == ignCol)
		  delete shape;
		
		//	add shape to component
		if (shape->attribute() == ShapeBasics::BACKGROUND)
		  //	register background regions as excluded
		  comp = &region[1];
		
		comp->push_back(shape);
	}

	return region;
} // Ds9::toRegion


std::ostream &operator<<(std::ostream &os, const ImageDisplayer::Region &r)
{
	os<<"# Region file format: DS9 version 3.0"<<endl;
	os<<"global color=white font=\"helvetica 10 normal\"";
	os<<" edit=1 move=1 delete=1 include=1 fixed=0"<<endl;
	for (unsigned i=0; i<r.size(); ++i) {
		const Component &c = r[i];
		for (unsigned n=0; n<c.size(); ++n) {
			bool commentStarted = false;
			switch (r.coordType()) {
				case ImageDisplayer::FK5:
					os<<"fk5;";
					break;
				case ImageDisplayer::IMAGE:
					os<<"image;";
					break;
				default:
					os << "physical;";
					break;
			}
			if ((!c.included() && c[n]->included()) ||
				(c.included() && !c[n]->included()))
				os<<'-';
			double rx = c[n]->refPoint().x(),
				   ry = c[n]->refPoint().y();
			switch(c[n]->type()) {
				case ShapeBasics::POINT:
					os<<"point("<<rx<<','<<ry<<") # point=circle";
					commentStarted = true;
					break;
				case ShapeBasics::CIRCLE:
					os<<"circle("<<rx<<','<<ry<<','<<c[n]->radii()[0]
					  <<')';
					break;
				case ShapeBasics::ELLIPSE:
					os<<"ellipse("<<rx<<','<<ry<<','<<c[n]->radii()[0]
					  <<','<<c[n]->radii()[1]<<','<<c[n]->rotAngles()[0]<<')';
					break;
				case ShapeBasics::ROTBOX:
					os<<"box("<<rx<<','<<ry<<','<<c[n]->radii()[0]
					  <<','<<c[n]->radii()[1]<<','<<c[n]->rotAngles()[0]<<')';
					break;
				case ShapeBasics::POLYGON: {
					os<<"polygon(";
					std::vector<float> x = c[n]->pointsX(),
								  y = c[n]->pointsY();
					os<<x[0]<<','<<y[0];
					for (unsigned n=1; n<x.size(); ++n)
						os<<','<<x[n]<<','<<y[n];
					os<<')';
					break;
				}
				default:
					errstr<<"Shape `"<<c[n]->name()<<"' is not supported by "
							"ds9 - will be skipped"
						  <<Sas::warning("UnsupportedShape");
					break;
			}
			//	write optional attributes
			if (!commentStarted)
				os<<" #";
			os<<" color="<<ShapeBasics::toString(c[n]->color());
			if (c[n]->isLabeled())
				os<<" text={"<<c[n]->label()<<"}";
			os<<endl;
		}
	}

	return os;
} // operator<<


std::istream &operator>>(std::istream &is, ImageDisplayer::Region &region)
{
	std::string s;
	std::vector<std::string> v;
	while (is>>s)
		v.push_back(s);

	region = Ds9::toRegion(v);

	return is;
} // operator>>
